# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`structures` provides general data structures to help manage 
various objects for Kepware's configuration
"""
from enum import Enum

class KepServiceResponse:
    '''A class to represent a return object when calling a "service" API of Kepware. This is
    used to return the responses when a "service" is executed appropriately

    :param code: HTTP code returned

    :param message: return from the "service" call

    :param href: URL reference to the JOB that is created by the service API
    '''

    def __init__(self, code: str = '', message: str = '', href: str = ''):
        self.code = code
        self.message = message
        self.href = href
    
    def __str__(self):
        return '{"code": %s, "message": %s, "href": %s}' % (self.code, self.message, self.href)

class KepServiceStatus:
    '''A class to represent a status object when checking on a "service" API job state in Kepware. This is
    used to return the status of a "service" job

    :param complete: Boolean of service job completion status

    :param status: Status code of job

    :param message: Error message if service job fails
    
    '''
    def __init__(self, complete: bool = False, status: str = '', message: str = ''):
        self.status = status
        self.message = message
        self.complete = complete
    
    def __str__(self):
        return '{"complete": %s, "status": %s, "message": %s}' % (self.complete, self.status, self.message)

class _HttpDataAbstract:
    def __init__(self):
        self.payload = ''
        self.code = ''
        self.reason = ''

class FilterModifierEnum(Enum):
    '''Enum class to represent the various filter types that can be used in the Kepware API'''
    EQUAL = "eq"
    NOTEQUAL = "neq"
    GREATERTHAN = "gt"
    LESSTTHAN = "lt"
    GREATERTHANEQUAL = "gte"
    LESSTHANEQUAL = "lte"
    CONTAINS = "contains"
    NOTCONTAINS = "ncontains"
    STARTSWITH = "starts_with"
    NOTSTARTSWITH = "nstarts_with"
    ENDSWITH = "ends_with"
    NOTENDSWITH = "nends_with"

class FilterFieldEnum(Enum):
    ID = "id"
    TIMESTAMP = "timestamp"
    ACTION = "action"
    USER = "user"
    INTERFACE = "interface"
    DETAILS = "details"
    DATA = "data"

class Filter:
    '''A class to represent a filter object when calling the Kepware API. This is used to
    filter the results of a GET request for Audit Logs.

    :param name: Name of the object to filter on

    :param type: Type of filter to apply

    :param value: Value to filter on
    '''
    def __init__(self, field: FilterFieldEnum = FilterFieldEnum.ID, modifier: FilterModifierEnum = FilterModifierEnum.EQUAL, value: str = ''):
        self.field = field
        self.modifier = modifier
        self.value = value
    
    def __str__(self):
        return '{"field": %s, "modifier": %s, "value": %s}' % (self.field, self.modifier, self.value)



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`error` Exception classes raised by Kepconfig.
Includes KepError, KepURLError and KepHTTPError
"""

__all__ = ['KepError', 'KepURLError', 'KepHTTPError']


class KepError(Exception):
    '''General Exception class for Kepconfig.
    
    :param msg: General error message returned in string format.
    '''
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
            return 'KepError Error: %s' % (self.msg)

class KepURLError(KepError):
    '''Exception class raised by Kepconfig that derives responses from the urllib URLError exceptions.

    :param url: full url path of the request that flagged a URLError exception
    :param msg: reason parameter in URLError exception
    '''
    def __init__(self, url=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url
    
    @property
    def reason(self):
        return self.msg

    def __str__(self):
        return '<urlopen error %s>' % self.reason

class KepHTTPError(KepError):
    '''Exception class raised by Kepconfig that derives responses from the urllib HTTPError 
    exceptions. This exception class is also a valid HTTP response instance.  It behaves 
    this way because HTTP protocol errors are valid responses, with a status 
    code, headers, and a body.  In some contexts, an application may want to 
    handle an exception like a regular response.

    :param url: url parameter in HTTPError exception
    :param code: HTTP response code parameter in HTTPError exception
    :param hdrs: hdrs parameter in HTTPError exception
    :param payload: string of HTTP response payload that flagged HTTPError exception 
    :param msg: msg parameter in HTTPError exception
    '''
    def __init__(self, url=None, code=None, hdrs=None, payload=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url
        self.code = code
        self.hdrs = hdrs
        self.payload = payload
    
    @property
    def reason(self):
        return self.msg

    def __str__(self):
        return 'HTTP Error %s: %s' % (self.code, self.msg)


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`device` exposes an API to allow modifications (add, delete, modify) to 
device objects within the Kepware Configuration API
"""

from ..connection import KepServiceResponse, server
from ..error import KepHTTPError, KepError
from ..utils import _url_parse_object, path_split
from typing import Union
from . import channel, tag
import inspect

DEVICE_ROOT = '/devices'
ATG_URL = '/services/TagGeneration'

def _create_url(device = None):
    '''Creates url object for the "device" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure
    
    Returns the device specific url when a value is passed as the device name.
    '''
    if device == None:
        return DEVICE_ROOT
    else:
        return '{}/{}'.format(DEVICE_ROOT,_url_parse_object(device))

def add_device(server: server, channel_name: str, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"device"` or multiple `"device"` objects to a channel in Kepware. Can be used to pass children of a device object 
    such as tags and tag groups. This allows you to create a device and tags 
    all in one function, if desired.

    Additionally it can be used to pass a list of devices and it's children to be added all at once.

    :param server: instance of the `server` class
    :param channel_name: channel to add the device object(s)
    :param DATA: Dict or List of Dicts of the device(s) and it's children
    expected by Kepware Configuration API
    
    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    devices added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_add(server.url + channel._create_url(channel_name) + _create_url(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_device(server: server, device_path: str) -> bool:
    '''Delete a `"device"` object in Kepware. This will delete all children as well.

    :param server: instance of the `server` class
    :param device_path: path identifying device to delete. Standard Kepware address decimal notation string including the 
    device such as `"channel1.device1"`

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    path_obj = path_split(device_path)
    try:
        r = server._config_del(server.url + channel._create_url(path_obj['channel']) + _create_url(path_obj['device']))
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
    except KeyError as err:
        err_msg = 'Error: No {} identified in {} | Function: {}'.format(err,'device_path', inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)

def modify_device(server: server, device_path: str, DATA: dict, *, force: bool = False) -> bool:
    '''Modify a device object and it's properties in Kepware.

    :param server: instance of the `server` class
    :param device_path: path identifying device to modffy. Standard Kepware address decimal notation string including the 
    device such as `"channel1.device1"`
    :param DATA: Dict of the `device` properties to be modified
    :param force: *(optional)* if True, will force the configuration update to the Kepware server
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    device_data = server._force_update_check(force, DATA)

    path_obj = path_split(device_path)
    try:
        r = server._config_update(server.url + channel._create_url(path_obj['channel']) + _create_url(path_obj['device']), device_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
    except KeyError as err:
            err_msg = 'Error: No {} identified in {} | Function: {}'.format(err,'device_path', inspect.currentframe().f_code.co_name)
            raise KepError(err_msg)

def get_device(server: server, device_path: str) -> dict:
    '''Returns the properties of the device object.

    :param server: instance of the `server` class
    :param device_path: path identifying device to retrieve properties. Standard Kepware address decimal notation string including the 
    device such as `"channel1.device1"`

    :return: Dict of data for the device requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    path_obj = path_split(device_path)
    try:
        r = server._config_get(server.url + channel._create_url(path_obj['channel']) + _create_url(path_obj['device']))
        return r.payload
    except KeyError as err:
        err_msg = 'Error: No {} identified in {} | Function: {}'.format(err,'device_path', inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    # except Exception as err:
    #     print('Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(err)))
    #     raise err

def get_all_devices(server: server, channel_name: str, *, options: dict = None) -> list:
    '''Returns list of all device objects and their properties within a channel. Returned object is JSON list.
    
    :param server: instance of the `server` class
    :param channel: name of channel
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of devices. Options are `filter`, 
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`

    :return: List of data for all devices within the channel

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(f'{server.url}{channel._create_url(channel_name)}{_create_url()}', params= options)
    return r.payload

def auto_tag_gen(server: server, device_path: str, job_ttl: int = None) -> KepServiceResponse:
    '''Executes Auto Tag Generation function on devices that support the feature in Kepware
    
    :param server: instance of the `server` class
    :param device_path: path identifying device to modffy. Standard Kepware address decimal notation string including the 
    device such as `"channel1.device1"`

    :param job_ttl: *(optional)* Determines the number of seconds a job instance will exist following completion.

    :return: `KepServiceResponse` instance with job information
    
    :raises KepHTTPError: If urllib provides an HTTPError (If not HTTP code 202 [Accepted] or 429 [Too Busy] returned)
    :raises KepURLError: If urllib provides an URLError
    '''
    
    path_obj = path_split(device_path)
    try:
        url = server.url +channel._create_url(path_obj['channel']) + _create_url(path_obj['device']) + ATG_URL
        job = server._kep_service_execute(url, None, job_ttl)
        return job
    except KeyError as err:
        err_msg = 'Error: No {} identified in {} | Function: {}'.format(err,'device_path', inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)

def get_all_tags_tag_groups(server: server, device_path: str) -> dict:
    '''Returns the properties of all `"tag"` and `"tag group"` objects for as specific
    device in Kepware. Returned object is a dict of tag list and tag group list.

    The returned object resembles the example below, nested based on how many 
    levels the tag_group namespace has tags or tag_groups:

    Example return:

        {
            'tags': [tag1_dict, tag2_dict,...],
            'tag_groups':[
                {
                    tag_group1_properties,
                    'tags': [tag1_dict, tag2_dict,...]
                    'tag_groups':[sub_group1, subgroup2,...]
                }, 
                {
                    tag_group2_properties,
                    'tags': [tag1_dict, tag2_dict,...]
                    'tag_groups':[sub_group1, subgroup2,...]
                },...]
        } 

    :param server: instance of the `server` class
    :param device_path: path identifying device to modffy. Standard Kepware address decimal notation string including the 
    device such as `"channel1.device1"`

    :return: Dict of data for the tag structure for device requested at `"device_path"` location

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = tag.get_full_tag_structure(server, device_path,recursive=True)
    return r

def get_device_structure(server, device_path) -> dict:
    '''Returns the properties of `"device"` and includes all `"tag"` and `"tag group"` objects for as specific
    device in Kepware. Returned object is a dict of device properties including a tag list and tag group list.

    The returned object resembles example below, nested based on how many 
    levels the tag_group namespace has tags or tag_groups:

    Example return:
    
        {
            device_properties,
            'tags': [tag1_dict, tag2_dict,...],
            'tag_groups':[
                {
                    tag_group1_properties,
                    'tags': [tag1_dict, tag2_dict,...]
                    'tag_groups':[sub_group1, subgroup2,...]
                }, 
                {
                    tag_group2_properties,
                    'tags': [tag1_dict, tag2_dict,...]
                    'tag_groups':[sub_group1, subgroup2,...]
                },...]
        } 

    :param server: instance of the `server` class
    :param device_path: path identifying device to modffy. Standard Kepware address decimal notation string including the 
    device such as `"channel1.device1"`

    :return: Dict of data for the device structure at `"device_path"` location

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    tags = tag.get_full_tag_structure(server, device_path,recursive=True)
    device_properties = get_device(server,device_path)
    return {**device_properties, **tags}



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`profile` exposes an API to allow modifications (add, delete, modify) to 
profile objects for the UDD Profile Library plug-in within the Kepware Configuration API
"""

from ...connection import server
from ...error import KepHTTPError, KepError
from typing import Union

PROFILE_ROOT = '/project/_profile_library/profiles'

def add_profile(server: server, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"profile"` or a list of `"profile"` objects to the UDD Profile Library plug-in for Kepware. 

    :param server: instance of the `server` class
    :param DATA: Dict or List of Dicts of the profiles to add to the Profile Library 
    through Kepware Configuration API

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    profiles added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_add(f'{server.url}{PROFILE_ROOT}', DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_profile(server: server, profile_name: str) -> bool:
    '''Delete a `"profile"` object in UDD Profile Library plug-in for Kepware.
    
    :param server: instance of the `server` class
    :param profile_name: name of profile
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_del(f'{server.url}{PROFILE_ROOT}/{profile_name}')
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_profile(server: server, DATA: dict, profile_name: str = None, force: bool = False) -> bool:
    '''Modify a `"profile"` object and it's properties in Kepware. If a `"profile_name"` is not provided as an input,
    you need to identify the profile in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the profile that is to be modified.

    :param server: instance of the `server` class
    :param DATA: Dict or List of Dicts of the profile properties to be modified.
    :param profile_name: *(optional)* name of profile to modify. Only needed if not existing in `"DATA"`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    profile_data = server._force_update_check(force, DATA)
    if profile_name == None:
        try:
            r = server._config_update(f"{server.url}{PROFILE_ROOT}/{profile_data['common.ALLTYPES_NAME']}", profile_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No profile identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
    else:
        r = server._config_update(f'{server.url}{PROFILE_ROOT}/{profile_name}', profile_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_profile(server: server, profile_name: str = None, *, options: dict = None) -> Union[dict, list]:
    '''Returns the properties of the profile object or a list of all profiles and their 
    properties. Will return a list if `"profile_name"` is not provided.
    
    INPUTS:

    :param server: instance of the `server` class
    :param profile_name: *(optional)* name of profile. If not defined, will get all profiles
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate when getting a list of profiles. Options are `filter`, 
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`. Only used when profile_name is not defined.
    
    :return: Dict of the profile properties or List of Dicts for all profiles and their properties in the Profile Library

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    if profile_name == None:
        r = server._config_get(f'{server.url}{PROFILE_ROOT}', params= options)
    else:
        r = server._config_get(f'{server.url}{PROFILE_ROOT}/{profile_name}')
    return r.payload

def get_all_profiles(server: server, *, options: dict = None):
    '''Returns list of all profile objects and their properties. Returned object is JSON list.
    
    :param server: instance of the `server` class
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate when getting a list of profiles. Options are `filter`, 
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`. Only used when profile_name is not defined.
    
    :return: List of data for all profiles and their properties in the Profile Library

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    return get_profile(server, options= options)


# -------------------------------------------------------------------------
# Copyright (c) 2023 PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`client` exposes an API to allow modifications (add, delete, modify) to 
UA Gateway plug-in client connection objects within the Kepware Configuration API. 
Certificate store read, remove and trust functionality is also available for the 
client connections.
"""

from typing import Union
from ..connection import server
from ..error import KepError, KepHTTPError
from ..ua_gateway.common import _INTER_TYPE, _change_cert_trust,  _create_url_cert, _create_url_client, _delete_cert_truststore

def get_ua_client_connection(server: server, ua_client_connection: str) -> dict:
    '''Returns the properties of the UAG client connection object.
    
    :param server: instance of the `server` class
    :param ua_client_connection: name of ua_client_connection
    
    :return: Dict of data for the client connection requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url_client(ua_client_connection))
    return r.payload

def get_all_ua_client_connections(server: server,  *, options: dict = None) -> dict:
    '''Returns list of all UAG client connection objects and their properties.
    
    :param server: instance of the `server` class
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of client connections. Options are `filter`, 
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`
    
    :return: List of data for all client connections in Kepware server UAG

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url_client())
    return r.payload

def add_ua_client_connection(server: server, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"UAG client connection"` or multiple `"UAG client connection"` objects to Kepware. This allows you 
    to create a client connection with all needed properties.

    Additionally it can be used to pass a list of UAG client connections to be added all at once.

    :param server: instance of the `server` class
    :param DATA: Dict of the connection or a list of connections
    expected by Kepware Configuration API

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    connections added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_add(server.url + _create_url_client(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
    
def modify_ua_client_connection(server: server, DATA: dict, *, ua_client_connection: str = None, force: bool = False) -> bool:
    '''Modify a UAG client connection object and it's properties in Kepware. If a `"ua_client_connection"` is not provided as an input,
    you need to identify the client connection in the *'common.ALLTYPES_NAME'* property field in `"DATA"`. It will 
    assume that is the client connection that is to be modified.

    :param server: instance of the `server` class
    :param DATA: Dict of the `ua_client_connection` properties to be modified
    :param ua_client_connection: *(optional)* name of client connection to modify. Only needed if not existing in `"DATA"`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    client_data = server._force_update_check(force, DATA)
    if ua_client_connection == None:
        try:
            r = server._config_update(server.url + _create_url_client(client_data['common.ALLTYPES_NAME']), client_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No Channel identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except Exception as e:
        #     return 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
    else:
        r = server._config_update(server.url + _create_url_client(ua_client_connection), client_data)
        if r.code == 200: return True 
        else: 
            raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        
def del_ua_client_connection(server: server, ua_client_connection: str) -> bool:
    '''Delete a `"UAG client connection"` object in Kepware.
    
    :param server: instance of the `server` class
    :param ua_client_connection: name of client connection
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_del(server.url + _create_url_client(ua_client_connection))
    if r.code == 200: return True 
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_certificate(server: server, certificate: str) -> dict:
    '''Returns the properties of the UAG client connection certificate object in the UAG client connection 
    certificate store. These are UA server instance certificates that are used by UAG client connections for
    trust purposes in the UA security model.
    
    :param server: instance of the `server` class
    :param certificate: name of certificate
    
    :return: Dict of data for the certificate requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url_cert(_INTER_TYPE.CLIENT, certificate))
    return r.payload

def get_all_certificates(server: server,  *, options: dict = None) -> list:
    '''Returns list of all UAG client connection certificate objects and their properties. These are UA server instance 
    certificates that are used by UAG client connections for trust purposes in the UA security model.
    
    :param server: instance of the `server` class
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of certificates. Options are `filter`, 
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`
    
    :return: List of data for all certificates in Kepware server UAG client connection certificate store

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url_cert(_INTER_TYPE.CLIENT), params= options)
    return r.payload

def trust_certificate(server: server, certificate: str) -> bool:
    '''Trusts the certificate in the UAG client connection certifcate store. This is updating the trust state of UA server instance 
    certificates that are used by UAG client connections for trust purposes in the UA security model.

    :param server: instance of the `server` class
    :param certificate: name of certificate
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    return _change_cert_trust(server, _INTER_TYPE.CLIENT, certificate, True)
    
def reject_certificate(server: server, certificate: str) -> bool:
    '''Rejects the certificate in the UAG client connection certifcate store. This is updating the trust state of UA server instance 
    certificates that are used by UAG client connections for trust purposes in the UA security model.

    :param server: instance of the `server` class
    :param certificate: name of certificate
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    return _change_cert_trust(server, _INTER_TYPE.CLIENT, certificate, False)

def delete_certificate(server: server, certificate: str) -> bool:
    '''Deletes the certificate in the UAG client endpoint certificate store.

    :param server: instance of the `server` class
    :param certificate: name of certificate
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    return _delete_cert_truststore(server, _INTER_TYPE.CLIENT, certificate)


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`mapping` exposes an API to allow modifications (add, delete, modify) to 
column mapping objects in a Datalogger log group within the Kepware Configuration API
"""

from . import log_group as Log_Group
from ..error import KepError, KepHTTPError
from ..connection import server
from ..utils import _url_parse_object

MAPPING_ROOT = '/column_mappings'

def _create_url(mapping = None):
    '''Creates url object for the "column_mappings" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the mapping specific url when a value is passed as the column_mapping name.
    '''

    if mapping == None:
        return '{}'.format(MAPPING_ROOT)
    else:
        return '{}/{}'.format(MAPPING_ROOT, _url_parse_object(mapping))

def modify_mapping(server: server, log_group: str, DATA: dict, *, mapping: str = None, force: bool = False) -> bool:
    '''Modify a column `"mapping"` object and it's properties in Kepware. If a `"mapping"` is not provided as an input,
    you need to identify the column mapping in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the column mapping that is to be modified.

    :param server: instance of the `server` class
    :param log_group: name of log group for the mapping
    :param DATA: Dict of the mapping properties to be modified.
    :param mapping: *(optional)* column mapping to modify in the log group. Only needed if not existing in `"DATA"`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    mapping_data = server._force_update_check(force, DATA)
    
    if mapping == None:
        try:
            r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(mapping_data['common.ALLTYPES_NAME']), mapping_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No column mapping identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
    else:
        r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(mapping), mapping_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_mapping(server: server, log_group: str, mapping: str) -> dict:
    '''Returns the properties of the `"mapping"` object.
    
    :param server: instance of the `server` class
    :param log_group: name of log group for the mapping
    :param mapping: name of column mapping to retrieve properties
    
    :return: Dict of properties for the mapping object requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + Log_Group._create_url(log_group) + _create_url(mapping))
    return r.payload

def get_all_mappings(server: server, log_group: str, *, options: dict = None) -> list:
    '''Returns the properties of all column `"mapping"` objects for a log group.
    
    :param server: instance of the `server` class
    :param log_group: name of log group for the mapping
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of mapping items. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize'. Only used when exchange_name is not defined.

    :return: list of properties for all mapping items in the log group requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(f'{server.url}{Log_Group._create_url(log_group)}{_create_url()}', params= options)
    return r.payload



# -------------------------------------------------------------------------
# Copyright (c), PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`ua_server` exposes an API to allow modifications (add, delete, modify) to 
OPC UA Server endpoints within the Kepware Administration through the Kepware Configuration API
"""
from typing import Union
from ..error import KepHTTPError, KepError
from ..connection import server
from ..utils import _url_parse_object


UA_ROOT = '/admin/ua_endpoints'

def _create_url(endpoint = None):
    '''Creates url object for the "server_users" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the user specific url when a value is passed as the user name.
    '''
    
    if endpoint == None:
        return UA_ROOT
    else:
        return '{}/{}'.format(UA_ROOT, _url_parse_object(endpoint))

def add_endpoint(server: server, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add an `"endpoint"` or multiple `"endpoint"` objects to Kepware UA Server by passing a 
    list of endpoints to be added all at once.

    :param server: instance of the `server` class
    :param DATA: Dict or List of Dicts of the UA Endpoints to add

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    endpoints added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_endpoint(server: server, endpoint: str) -> bool:
    '''Delete an `"endpoint"` object in Kepware UA Server
    
    :param server: instance of the `server` class
    :param endpoint: name of endpoint to delete
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(endpoint))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_endpoint(server: server, DATA: dict, endpoint: str = None) -> bool:
    '''Modify a `"endpoint"` object and it's properties in Kepware UA Server. If a `"endpoint"` is not provided as an input,
    you need to identify the endpoint in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the endpoint that is to be modified.

    :param server: instance of the `server` class
    :param DATA: Dict of the UA endpoint properties to be modified.
    :param endpoint: *(optional)* name of endpoint to modify. Only needed if not existing in `"DATA"`
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    if endpoint == None:
        try:
            r = server._config_update(server.url + _create_url(DATA['common.ALLTYPES_NAME']), DATA)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No UA Endpoint identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
    else:
        r = server._config_update(server.url + _create_url(endpoint), DATA)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_endpoint(server: server, endpoint: str) -> dict:
    '''Returns the properties of the `"endpoint"` object.
    
    :param server: instance of the `server` class
    :param endpoint: name of endpoint to retrieve
    
    :return: Dict of properties for the UA endpoint requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_get(server.url + _create_url(endpoint))
    return r.payload

def get_all_endpoints(server: server, *, options: dict = None) -> list:
    '''Returns list of all `"endpoint"` objects and their properties.
    
    :param server: instance of the `server` class
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of UA endpoints. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize.
    
    :return: List of properties for all UA endpoints requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_get(f'{server.url}{_create_url()}', params= options)
    return r.payload


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`agent` exposes an API to allow modifications (add, delete, modify) to 
Iot Gateway agent objects within the Kepware Configuration API
"""

# from .. import connection 
from typing import Union
from ..connection import server
from .. import iot_gateway as IOT
from ..error import KepError, KepHTTPError
import inspect
from ..utils import _url_parse_object

IOT_ROOT_URL = '/project/_iot_gateway'
MQTT_CLIENT_URL = '/mqtt_clients'
REST_CLIENT_URL = '/rest_clients'
REST_SERVER_URL = '/rest_servers'
THINGWORX_URL = '/thingworx_clients'

def _create_url(agent_type, agent = None):
    '''Creates url object for the "agent" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the agent specific url when a value is passed as the agent name.
    '''

    if agent == None:
        if agent_type == IOT.MQTT_CLIENT_AGENT:
            return '{}{}'.format(IOT_ROOT_URL, MQTT_CLIENT_URL)
        elif agent_type == IOT.REST_CLIENT_AGENT:
            return '{}{}'.format(IOT_ROOT_URL, REST_CLIENT_URL)
        elif agent_type == IOT.REST_SERVER_AGENT:
            return '{}{}'.format(IOT_ROOT_URL, REST_SERVER_URL)
        elif agent_type == IOT.THINGWORX_AGENT:
            return '{}{}'.format(IOT_ROOT_URL, THINGWORX_URL)
        else:
            pass
    else:
        if agent_type == IOT.MQTT_CLIENT_AGENT:
            return '{}{}/{}'.format(IOT_ROOT_URL, MQTT_CLIENT_URL, _url_parse_object(agent))
        elif agent_type == IOT.REST_CLIENT_AGENT:
            return '{}{}/{}'.format(IOT_ROOT_URL, REST_CLIENT_URL, _url_parse_object(agent))
        elif agent_type == IOT.REST_SERVER_AGENT:
            return '{}{}/{}'.format(IOT_ROOT_URL, REST_SERVER_URL,_url_parse_object(agent))
        elif agent_type == IOT.THINGWORX_AGENT:
            return '{}{}/{}'.format(IOT_ROOT_URL, THINGWORX_URL, _url_parse_object(agent))
        else:
            pass


def add_iot_agent(server: server, DATA: Union[dict, list], agent_type: str = None) -> Union[bool, list]:
    '''Add a  `"agent"` or multiple `"agent"` objects of a specific type to Kepware's IoT Gateway. Can be used to pass children of an
    agent object such as iot items. This allows you to create an agent and iot items if desired. Multiple Agents need to be of the 
    same type.

    Additionally it can be used to pass a list of agents and it's children to be added all at once.

    :param server: instance of the `server` class
    :param DATA: Dict or List of Dicts of the agent and it's children
    expected by Kepware Configuration API
    :param agent_type: *(optional)* agent type to add to IoT Gateway. Only needed if not existing in `"DATA"`. Valid values are 
    `MQTT Client`, `REST Client` or `REST Server`

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    iot agents added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    if agent_type == None:
        try:
            # If it's a list, use the first agents type
            if isinstance(DATA, list): agent_type = DATA[0]['iot_gateway.AGENTTYPES_TYPE']
            else: agent_type = DATA['iot_gateway.AGENTTYPES_TYPE']
        except KeyError as err:
            err_msg = 'Error: No agent identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
    
    r = server._config_add(server.url + _create_url(agent_type), DATA)
    if r.code == 201: return True 
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_iot_agent(server: server, agent: str, agent_type: str) -> bool:
    '''Delete a `"agent"` object in Kepware. This will delete all children as well
    
    :param server: instance of the `server` class
    :param agent: name of IoT Agent to delete
    :param agent_type: *(optional)* agent type to delete in IoT Gateway. Valid values are 
    `MQTT Client`, `REST Client` or `REST Server`

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_del(server.url + _create_url(agent_type, agent))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_iot_agent(server: server, DATA: dict, *, agent: str = None, agent_type: str = None, force: bool = False) -> bool:
    '''Modify a `"agent"` object and it's properties in Kepware. If a `"agent"` is not provided as an input,
    you need to identify the agent in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the agent that is to be modified.

    :param server: instance of the `server` class
    :param DATA: Dict of the iot agent properties to be modified.
    :param agent: *(optional)* name of IoT agent to modify. Only needed if not existing in `"DATA"`
    :param agent_type: *(optional)* agent type to modify. Only needed if not existing in `"DATA"`. Valid values are 
    `MQTT Client`, `REST Client` or `REST Server`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    agent_data = server._force_update_check(force, DATA)
    
    if agent_type == None:
        if 'iot_gateway.AGENTTYPES_TYPE' in DATA:
            agent_type = DATA['iot_gateway.AGENTTYPES_TYPE']
        else:
            err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, 'No Agent type defined.')
            raise KepError(err_msg)
    if agent == None:
        try:
            r = server._config_update(server.url + _create_url(agent_type, agent_data['common.ALLTYPES_NAME']), agent_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No agent identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
    else:
        r = server._config_update(server.url + _create_url(agent_type, agent), agent_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_iot_agent(server: server, agent: str, agent_type: str) -> dict:
    '''Returns the properties of the `"agent"` object.
    
    :param server: instance of the `server` class
    :param DATA: Dict of the iot agent properties to be modified.
    :param agent: name of IoT agent to retrieve
    :param agent_type: agent type. Valid values are `MQTT Client`, `REST Client` or `REST Server`

    :return: Dict of properties for the iot agent requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url(agent_type, agent))
    return r.payload

def get_all_iot_agents(server: server, agent_type: str, *, options: dict = None) -> list:
    '''Returns the properties of all `"agent"` objects for a specific agent type. Returned object is JSON list.
    
    :param server: instance of the `server` class
    :param agent_type: agent type. Valid values are `MQTT Client`, `REST Client` or `REST Server`
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of IoT agents. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize'. Only used when exchange_name is not defined.

    :return: list of properties for all IoT agents requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(f'{server.url}{_create_url(agent_type)}', params= options)
    return r.payload



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`ranges` exposes an API to allow modifications (add, delete, modify) to 
range objects in exchanges for EGD devices within the Kepware Configuration API
"""

from typing import Union
from .. import egd as EGD
from ...connection import server
from ...error import KepError, KepHTTPError
from ...utils import _url_parse_object

RANGES_ROOT = '/ranges'

def _create_url(device_path, ex_type, exchange_name, range = None):
    '''Creates url object for the "range" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the range specific url when a value is passed as the range name.
    '''
    exchange_root = EGD.exchange._create_url(device_path, ex_type, exchange_name)

    if range == None:
        return '{}{}'.format(exchange_root, RANGES_ROOT)
    else:
        return '{}{}/{}'.format(exchange_root, RANGES_ROOT, _url_parse_object(range))

def add_range(server: server, device_path: str, ex_type: str, exchange_name: str, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"range"` or multiple `"range"` objects to Kepware. This allows you to 
    create a range or multiple ranges all in one function, if desired.

    When passing multiple ranges, they will be populated in the same order
    in the list sent. Ensure you provide the list in the order desired.

    :param server: instance of the `server` class
    :param device_path: path to EGD device. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param ex_type: type of exchange, either `CONSUMER` or `PRODUCER`
    :param exchange_name: name of exchange that range is located
    :param DATA: Dict or List of Dicts of the range(s) to add

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    ranges added that failed.
        
    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(device_path, ex_type, exchange_name), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_range(server: server, device_path: str, ex_type: str, exchange_name: str, range_name: str) -> bool:
    '''Delete a `"range"` object in Kepware.
    
    :param server: instance of the `server` class
    :param device_path: path to EGD device. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param ex_type: type of exchange, either `CONSUMER` or `PRODUCER`
    :param exchange_name: name of exchange that range is located
    :param range_name: name of range to delete
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(device_path, ex_type, exchange_name, range_name))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_range(server: server, device_path: str, ex_type: str, exchange_name: str, DATA: dict, *, range_name: str = None, force: bool = False) -> bool:
    '''Modify a `"range"` object and it's properties in Kepware. If a `"range_name"` is not provided as an input,
    you need to identify the range in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the range that is to be modified.

    :param server: instance of the `server` class
    :param device_path: path to EGD device. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param ex_type: type of exchange, either `CONSUMER` or `PRODUCER`
    :param exchange_name: name of exchange that range is located
    :param DATA: Dict of the range properties to be modified.
    :param range_name: *(optional)* name of range to to modify. Only needed if not existing in `"DATA"`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    range_data = server._force_update_check(force, DATA)
    if range_name == None:
        try:
            r = server._config_update(server.url + _create_url(device_path, ex_type, exchange_name, range_data['common.ALLTYPES_NAME']), range_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No range identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
    else:
        r = server._config_update(server.url + _create_url(device_path, ex_type, exchange_name, range_name), range_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_range(server: server, device_path: str, ex_type: str, exchange_name: str, range_name: str = None, *, options: dict = None) -> Union[dict, list]:
    '''Returns the properties of the `"range"` object or a list of all ranges.
    
    :param server: instance of the `server` class
    :param device_path: path to EGD device. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param ex_type: type of exchange, either `CONSUMER` or `PRODUCER`
    :param exchange_name: name of exchange that range is located
    :param DATA: Dict of the range properties to be modified.
    :param range_name: *(optional)* name of range to retrieve. If not defined, get all ranges
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of exchanges. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize'. Only used when range_name is not defined.
    
    :return: Dict of properties for the range requested or a List of ranges and their properties

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    if range_name == None:
        r = server._config_get(f'{server.url}{_create_url(device_path, ex_type, exchange_name)}', params= options)
    else:
        r = server._config_get(f'{server.url}{_create_url(device_path, ex_type, exchange_name, range_name)}')
    return r.payload


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`triggers` exposes an API to allow modifications (add, delete, modify) to 
trigger objects in a Datalogger log group within the Kepware Configuration API
"""

from typing import Union
from . import log_group as Log_Group
from ..error import KepError, KepHTTPError
from ..connection import server
from ..utils import _url_parse_object

TRIGGERS_ROOT = '/triggers'

def _create_url(trigger = None):
    '''Creates url object for the "trigger" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the trigger specific url when a value is passed as the trigger name.
    '''

    if trigger == None:
        return '{}'.format(TRIGGERS_ROOT)
    else:
        return '{}/{}'.format(TRIGGERS_ROOT, _url_parse_object(trigger))


def add_trigger(server: server, log_group: str, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"trigger"` or multiple `"trigger"` objects to a log group in Kepware's Datalogger. It can 
    be used to pass a list of triggers to be added all at once.

    :param server: instance of the `server` class
    :param log_group: name of log group for the trigger items
    :param DATA: Dict or a list of the trigger items to add through Kepware Configuration API

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    triggers added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_add(server.url + Log_Group._create_url(log_group) + _create_url(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_trigger(server: server, log_group: str, trigger: str) -> bool:
    '''Delete a `"trigger"` object of a log group in Kepware's Datalogger.
    
    :param server: instance of the `server` class
    :param log_group: name of log group for the trigger items
    :param trigger: name of trigger to delete

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_del(server.url + Log_Group._create_url(log_group) + _create_url(trigger))
    if r.code == 200: return True
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_trigger(server: server, log_group: str, DATA: dict, *, trigger: str = None, force: bool = False)  -> bool:
    '''Modify a `"trigger"` object and it's properties in Kepware. If a `"trigger"` is not provided as an input,
    you need to identify the trigger in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the trigger that is to be modified.

    :param server: instance of the `server` class
    :param log_group: name of log group for the trigger items
    :param DATA: Dict of the trigger properties to be modified.
    :param trigger: *(optional)* name of trigger to modify in the log group. Only needed if not existing in `"DATA"`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    trigger_data = server._force_update_check(force, DATA)
    
    if trigger == None:
        try:
            r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(trigger_data['common.ALLTYPES_NAME']), trigger_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No trigger identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except:
        #     return 'Error: Error with {}'.format(inspect.currentframe().f_code.co_name)
    else:
        r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(trigger), trigger_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_trigger(server, log_group, trigger) -> dict:
    '''Returns the properties of the `"trigger"` object.
    
    :param server: instance of the `server` class
    :param log_group: name of log group for the trigger items
    :param trigger: name of trigger to retrieve

    :return: Dict of properties for the trigger requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + Log_Group._create_url(log_group) + _create_url(trigger))
    return r.payload

def get_all_triggers(server: server, log_group: str, *, options: dict = None) -> list:
    '''Returns the properties of all `"trigger"` objects for a log group.
    
    :param server: instance of the `server` class
    :param log_group: name of log group for the trigger items

    :return: Dict of properties for the trigger requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of triggers. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize'. Only used when exchange_name is not defined.

    :return: list of properties for all triggers in the log group requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(f'{server.url}{Log_Group._create_url(log_group)}{_create_url()}', params= options)
    return r.payload



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`log_group` exposes an API to allow modifications (add, delete, modify) to 
log group objects in DataLogger within the Kepware Configuration API
"""
from typing import Union
from ..connection import KepServiceResponse, server
from ..error import KepError, KepHTTPError
from ..utils import _url_parse_object

ENABLE_PROPERTY = 'datalogger.LOG_GROUP_ENABLED'
LOG_GROUP_ROOT = '/project/_datalogger/log_groups'
SERVICES_ROOT = '/services'
def _create_url(log_group = None):
    '''Creates url object for the "log_group" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the agent specific url when a value is passed as the agent name.
    '''

    if log_group == None:
        return '{}'.format(LOG_GROUP_ROOT)
    else:
        return '{}/{}'.format(LOG_GROUP_ROOT, _url_parse_object(log_group))


def add_log_group(server: server, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"log group"` or multiple `"log groups"` objects to Kepware's DataLogger. It can be used 
    to pass a list of log groups to be added all at once.

    :param server: instance of the `server` class
    :param DATA: Dict or a list of the log groups to add through Kepware Configuration API

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    log groups added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(), DATA)
    if r.code == 201: return True 
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_log_group(server: server, log_group: str) -> bool:
    '''Delete a `"log group"` object in Kepware's Datalogger.
    
    :param server: instance of the `server` class
    :param log_group: name of log group to delete

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_del(server.url + _create_url(log_group))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_log_group(server: server, DATA: dict, *, log_group: str = None, force: bool = False) -> bool:
    '''Modify a `"log group"` object and it's properties in Kepware's Datalogger. If a `"log group"` is not provided as an input,
    you need to identify the log group in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the log group that is to be modified.

    :param server: instance of the `server` class
    :param DATA: Dict of the log group properties to be modified.
    :param log_group: *(optional)* name of log group to modify. Only needed if not existing in `"DATA"`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    log_group_data = server._force_update_check(force, DATA)
    
    if log_group == None:
        try:
            r = server._config_update(server.url + _create_url(log_group_data['common.ALLTYPES_NAME']), log_group_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No log group identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
    else:
        r = server._config_update(server.url + _create_url(log_group), log_group_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_log_group(server: server, log_group: str) -> dict:
    '''Returns the properties of the `"log group"` object.
    
    :param server: instance of the `server` class
    :param log_group: name of log group to retrieve

    :return: Dict of properties for the log group requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url(log_group))
    return r.payload

def get_all_log_groups(server: server, *, options: dict = None) -> list:
    '''Returns the properties of all log group objects for Kepware's Datalogger. Returned object is JSON list.
    
    :param server: instance of the `server` class
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of log groups. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize'. Only used when exchange_name is not defined.

    :return: list of properties for all log groups requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(f'{server.url}{_create_url()}', params= options)
    return r.payload

def enable_log_group(server: server, log_group: str) -> bool:
    '''Enable the `"log group"`.
    
    :param server: instance of the `server` class
    :param log_group: name of log group to enable

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: True}
    return modify_log_group(server, DATA, log_group= log_group)

def disable_log_group(server: server, log_group: str) -> bool:
    '''Disable the log group. Returned object is JSON.
    
    :param server: instance of the `server` class
    :param log_group: name of log group to enable

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: False}
    return modify_log_group(server, DATA, log_group= log_group)

def reset_column_mapping_service(server: server, log_group: str, job_ttl: int = None) -> KepServiceResponse:
    '''Executes a ResetColumnMapping serivce call to the log group

    :param server: instance of the `server` class
    :param log_group: name of log group to enable
    :param job_ttl: *(optional)* Determines the number of seconds a job instance will exist following completion.

    :return: `KepServiceResponse` instance with job information
    
    :raises KepHTTPError: If urllib provides an HTTPError (If not HTTP code 202 [Accepted] or 429 [Too Busy] returned)
    :raises KepURLError: If urllib provides an URLError
    '''

    url = server.url + _create_url(log_group) + SERVICES_ROOT + '/ResetColumnMapping'
    job = server._kep_service_execute(url, None, job_ttl)
    return job


# -------------------------------------------------------------------------
# Copyright (c), PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`lls` exposes an API to allow modifications to Local License Server parameters in 
the Kepware Administration through the Kepware Configuration API
"""
from .. import connection
from typing import Union
from ..error import KepHTTPError, KepError



LLS_ROOT = '/admin'
FORCE_CHECK_URL = '/project/services/ForceLicenseCheck'
LICENSING_SERVER_PORT = "libadminsettings.LICENSING_SERVER_PORT"
LICENSING_SERVER_NAME = "libadminsettings.LICENSING_SERVER_NAME"
LICENSING_SERVER_ENABLE = "libadminsettings.LICENSING_SERVER_ENABLE"
LICENSING_CHECK_PERIOD_MINS = "libadminsettings.LICENSING_CHECK_PERIOD_MINS"
LICENSING_SERVER_SSL_PORT = "libadminsettings.LICENSING_SERVER_SSL_PORT"
LICENSING_SERVER_ALLOW_INSECURE_COMMS = "libadminsettings.LICENSING_SERVER_ALLOW_INSECURE_COMMS"
LICENSING_SERVER_ALLOW_SELF_SIGNED_CERTS = "libadminsettings.LICENSING_SERVER_ALLOW_SELF_SIGNED_CERTS"
LICENSING_CLIENT_ALIAS = "libadminsettings.LICENSING_CLIENT_ALIAS"

class lls_config:
    '''A class to represent a admin properties for the Local License Server connection from an instance of Kepware. 
    This object is used to easily manage the LLS parameters for a Kepware instance. 

    :param server_name: Host name or IP address of the LLS server
    :param server_port: HTTP/non-SSL port to target for the LLS server
    :param check_period: Period that Kepware checks licensing status
    :param server_port_SSL: HTTPS/SSL port to target for the LLS server
    :param allow_insecure_comms: When True, use HTTP/non-SSL connection to LLS
    :param allow_self_signed_certs: Allow for self signed certificates to be used during HTTPS/SSL connections to the LLS
    :param instance_alias_name: Alias name for LLS to use as reference to this Kepware instance
    '''

    def __init__(self, config = {}):
        self.server_name = config[LICENSING_SERVER_NAME] if LICENSING_SERVER_NAME in config else ''
        self.server_port = config[LICENSING_SERVER_PORT] if LICENSING_SERVER_PORT in config else 7070
        self.check_period = config[LICENSING_CHECK_PERIOD_MINS] if LICENSING_CHECK_PERIOD_MINS in config else 5
        self.server_port_SSL = config[LICENSING_SERVER_SSL_PORT] if LICENSING_SERVER_SSL_PORT in config else 1883
        self.allow_insecure_comms = config[LICENSING_SERVER_ALLOW_INSECURE_COMMS] if LICENSING_SERVER_ALLOW_INSECURE_COMMS in config else False
        self.allow_self_signed_certs = config[LICENSING_SERVER_ALLOW_SELF_SIGNED_CERTS] if LICENSING_SERVER_ALLOW_SELF_SIGNED_CERTS in config else False
        self.instance_alias_name = config[LICENSING_CLIENT_ALIAS] if LICENSING_CLIENT_ALIAS in config else ''
    
    def _get_dict(self):
        return {
            LICENSING_SERVER_PORT: self.server_port,
            LICENSING_SERVER_NAME: self.server_name,
            LICENSING_CHECK_PERIOD_MINS: self.check_period,
            LICENSING_SERVER_SSL_PORT: self.server_port_SSL,
            LICENSING_SERVER_ALLOW_INSECURE_COMMS: self.allow_insecure_comms,
            LICENSING_SERVER_ALLOW_SELF_SIGNED_CERTS: self.allow_self_signed_certs,
            LICENSING_CLIENT_ALIAS: self.instance_alias_name
        }
    
    def __str__(self) -> str:
        return "{}".format(self._get_dict())

def get_lls_config(server: connection.server) -> lls_config:
    '''Returns the properties of the Local License server connection properties. Returned object is `lls_config` class object.
    
    :param server: instance of the `server` class
    
    :return: `lls_config` class object with lls connection configuration

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_get(server.url + LLS_ROOT)
    return lls_config(r.payload)

def update_lls_config(server: connection.server, config: lls_config) -> bool:
    '''Updates the Local License Server connection properties for Kepware.
    
    :param server: instance of the `server` class
    :param config: `lls_config` class object with lls connection configuration
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    DATA = config._get_dict()
    r = server._config_update(server.url + LLS_ROOT, DATA)
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def enable_lls(server: connection.server) -> bool:
    '''Enables the Local License Server connection for Kepware.
    
    :param server: instance of the `server` class
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_update(server.url + LLS_ROOT, {LICENSING_SERVER_ENABLE: True})
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def disable_lls(server: connection.server) -> bool:
    '''Disables the Local License Server connection for Kepware.
    
    :param server: instance of the `server` class
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_update(server.url + LLS_ROOT, {LICENSING_SERVER_ENABLE: False})
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def force_license_check(server: connection.server, job_ttl: int = None):
    '''Executes a ForceLicenseCheck service call to the Kepware instance. This triggers the server to verify the 
    license state of the license received from the Local License Server.

    :param server: instance of the `server` class
    :param job_ttl: *(optional)* Determines the number of seconds a job instance will exist following completion.

    :return: `KepServiceResponse` instance with job information

    :raises KepHTTPError: If urllib provides an HTTPError (If not HTTP code 202 [Accepted] or 429 [Too Busy] returned)
    :raises KepURLError: If urllib provides an URLError
    '''

    url = f'{server.url}{FORCE_CHECK_URL}'
    job = server._kep_service_execute(url, None, job_ttl)
    return job


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`log_items` exposes an API to allow modifications (add, delete, modify) to 
log item (tag) objects in a Datalogger log group within the Kepware Configuration API
"""
from typing import Union
from . import log_group as Log_Group
from ..error import KepError, KepHTTPError
from ..connection import server
from ..utils import _url_parse_object

LOG_ITEMS_ROOT = '/log_items'

def _create_url(log_item = None):
    '''Creates url object for the "log_item" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the log_item specific url when a value is passed as the log_item name.
    '''

    if log_item == None:
        return '{}'.format(LOG_ITEMS_ROOT)
    else:
        return '{}/{}'.format(LOG_ITEMS_ROOT, _url_parse_object(log_item))


def add_log_item(server: server, log_group: str, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"log item"` or multiple `"log item"` objects to a log group in Kepware's Datalogger. It can 
    be used to pass a list of log items to be added all at once.

    :param server: instance of the `server` class
    :param log_group: name of log group that the log items will be added
    :param DATA: Dict or a list of the log items to add through Kepware Configuration API

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    log items added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_add(server.url + Log_Group._create_url(log_group) + _create_url(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_log_item(server: server, log_group: str, log_item: str) -> bool:
    '''Delete a `"log item"` object of a log group in Kepware's Datalogger.
    
    :param server: instance of the `server` class
    :param log_group: name of log group that log item exists
    :param log_item: name of log item to delete

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_del(server.url + Log_Group._create_url(log_group) + _create_url(log_item))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_log_item(server: server, log_group: str, DATA: dict, *, log_item: str = None, force: bool = False) -> bool:
    '''Modify a `"log_item"` object and it's properties in Kepware. If a `"log_item"` is not provided as an input,
    you need to identify the log_item in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the log_item that is to be modified.

    :param server: instance of the `server` class
    :param log_group: name of log group that log item exists
    :param DATA: Dict of the log item properties to be modified.
    :param log_item: *(optional)* name of log item to modify. Only needed if not existing in `"DATA"`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    log_item_data = server._force_update_check(force, DATA)
    
    if log_item == None:
        try:
            r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(log_item_data['common.ALLTYPES_NAME']), log_item_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg ='Error: No log item identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
    else:
        r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(log_item), log_item_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_log_item(server, log_group, log_item) -> dict:
    '''Returns the properties of the `"log item"` object.
    
    :param server: instance of the `server` class
    :param log_group: name of log group that log item exists
    :param log_item: name of log item to retrieve

    :return: Dict of properties for the log group requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + Log_Group._create_url(log_group) + _create_url(log_item))
    return r.payload

def get_all_log_items(server: server, log_group: str, *, options: dict = None) -> list:
    '''Returns the properties of all `"log item"` objects for a log group.
    
    :param server: instance of the `server` class
    :param log_group: name of log group that log item exists
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of log groups. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize'. Only used when exchange_name is not defined.

    :return: list of properties for all log items in the log group requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(f'{server.url}{Log_Group._create_url(log_group)}{_create_url()}', params= options)
    return r.payload



# -------------------------------------------------------------------------
# Copyright (c) 2023 PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`server` exposes an API to allow modifications (add, delete, modify) to 
UA Gateway plug-in server endpoint objects within the Kepware Configuration API. 
Certificate store read, remove and trust functionality is also available for the 
server endpoints.
"""

from typing import Union
from ..connection import server
from ..error import KepError, KepHTTPError
from ..ua_gateway.common import _INTER_TYPE, _change_cert_trust, _create_url_cert, _create_url_server, _delete_cert_truststore, SERVER_ROOT

def get_uag_server_interface_properties(server: server) -> dict:
    ''' Get the UAG Server Interface Properties of the Kepware instance. These properties expose User Identify
    Policy, Security Policy and other Communication properties that are applied across all UAG server endpoints.
        
    :return: Dict of all the UAG Server Interface properties

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_get(server.url + SERVER_ROOT)
    return r.payload

def modify_uag_server_interface_properties(server: server, DATA: dict, force: bool = False) -> bool:
    ''' Modify the UAG Server Interface Properties of the Kepware instance. These properties expose User Identify
    Policy, Security Policy and other Communication properties that are applied across all UAG server endpoints.

    :param DATA: Dict of the UAG Server Interface properties to be modified
    :param force: *(optional)* if True, will force the configuration update to the Kepware server
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    prop_data = server._force_update_check(force, DATA)
    r = server._config_update(server.url + SERVER_ROOT, prop_data)
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)


def get_ua_server_endpoint(server: server, ua_server_endpoint: str) -> dict:
    '''Returns the properties of the UAG server endpoint object.
    
    :param server: instance of the `server` class
    :param ua_server_endpoint: name of ua_server_endpoint
    
    :return: Dict of data for the server endpoint requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url_server(ua_server_endpoint))
    return r.payload

def get_all_ua_server_endpoints(server: server,  *, options: dict = None) -> dict:
    '''Returns list of all UAG server endpoint objects and their properties.
    
    :param server: instance of the `server` class
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of server endpoints. Options are `filter`, 
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`
    
    :return: List of data for all server endpoints in Kepware server UAG

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url_server())
    return r.payload

def add_ua_server_endpoint(server: server, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"UAG server endpoint"` or multiple `"UAG server endpoint"` objects to Kepware. This allows you 
    to create a server endpoint with all needed properties.

    Additionally it can be used to pass a list of UAG server endpoint to be added all at once.

    :param server: instance of the `server` class
    :param DATA: Dict of the endpoint or a list of endpoints
    expected by Kepware Configuration API

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    endpoints added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_add(server.url + _create_url_server(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
    
def modify_ua_server_endpoint(server: server, DATA: dict, *, ua_server_endpoint: str = None, force: bool = False) -> bool:
    '''Modify a UAG server endpoint object and it's properties in Kepware. If a `"ua_server_endpoint"` is not provided as an input,
    you need to identify the client connection in the *'common.ALLTYPES_NAME'* property field in `"DATA"`. It will 
    assume that is the client connection that is to be modified.

    :param server: instance of the `server` class
    :param DATA: Dict of the `ua_server_endpoint` properties to be modified
    :param ua_server_endpoint: *(optional)* name of server endpoint to modify. Only needed if not existing in `"DATA"`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    server_data = server._force_update_check(force, DATA)
    if ua_server_endpoint == None:
        try:
            r = server._config_update(server.url + _create_url_server(server_data['common.ALLTYPES_NAME']), server_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No Channel identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except Exception as e:
        #     return 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
    else:
        r = server._config_update(server.url + _create_url_server(ua_server_endpoint), server_data)
        if r.code == 200: return True 
        else: 
            raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        
def del_ua_server_endpoint(server: server, ua_server_endpoint: str) -> bool:
    '''Delete a `"UAG server endpoint"` object in Kepware.
    
    :param server: instance of the `server` class
    :param ua_server_endpoint: name of server endpoint
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_del(server.url + _create_url_server(ua_server_endpoint))
    if r.code == 200: return True 
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_certificate(server: server, certificate: str) -> dict:
    '''Returns the properties of the UAG server endpoint certificate object in the UAG client connection 
    certificate store. These are UA client instance certificates that are used by UAG server endpoints for
    trust purposes in the UA security model.
    
    :param server: instance of the `server` class
    :param certificate: name of certificate
    
    :return: Dict of data for the certificate requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url_cert(_INTER_TYPE.SERVER, certificate))
    return r.payload

def get_all_certificates(server: server, *, options: dict = None) -> list:
    '''Returns list of all UAG server endpoint certificate objects and their properties.This is updating the trust state of UA client instance 
    certificates that are used by UAG server endpoints for trust purposes in the UA security model. These are UA client instance certificates 
    that are used by UAG server endpoints for trust purposes in the UA security model.
    
    :param server: instance of the `server` class
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of certificates. Options are `filter`, 
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`
    
    :return: List of data for all certificates in Kepware server UAG server endpoint certificate store

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url_cert(_INTER_TYPE.SERVER), params= options)
    return r.payload

def trust_certificate(server: server, certificate: str) -> bool:
    '''Trusts the certificate in the UAG server endpoint certificate store. This is updating the trust state of UA client instance 
    certificates that are used by UAG server endpoints for trust purposes in the UA security model.

    :param server: instance of the `server` class
    :param certificate: name of certificate
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    return _change_cert_trust(server, _INTER_TYPE.SERVER, certificate, True)

def reject_certificate(server: server, certificate: str) -> bool:
    '''Rejects the certificate in the UAG server endpoint certificate store.

    :param server: instance of the `server` class
    :param certificate: name of certificate
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    return _change_cert_trust(server, _INTER_TYPE.SERVER, certificate, False)

def delete_certificate(server: server, certificate: str) -> bool:
    '''Deletes the certificate in the UAG server endpoint certificate store.

    :param server: instance of the `server` class
    :param certificate: name of certificate
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    return _delete_cert_truststore(server, _INTER_TYPE.SERVER, certificate)


# -------------------------------------------------------------------------
# Copyright (c) 2023 PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`ua_gateway` module provides support for Kepware's UA Gateway plug-in 
specific objects within the Kepware Configuration API
"""

from . import certificates, client, server


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`utils` provides general utilities to help manage 
various objects for Kepware's configuration
"""

from urllib import parse

def path_split(path: str):
    '''Used to split the standard Kepware address decimal notation into a dict that contains the 
    *channel*, *device* and *tag_path* keys.

    :param path: standard Kepware address in decimal notation ("ch1.dev1.tg1.tg2.tg3")
    :return: dict that contains the "channel", "device" and "tag_path"
    :rtype: dict

    Ex: path = "ch1.dev1.tg1.tg2.tg3"

    return = {'channel': 'ch1', 'device': 'dev1', 'tag_path': ['tg1','tg2','tg3']}

    Ex: path = "ch1.dev1"

    return = {'channel': 'ch1', 'device': 'dev1'}
    '''
    path_list = path.split('.', 2)
    path_obj = {}
    for x in range(0, len(path_list)):
        if x == 0:
            path_obj['channel'] = path_list[0]
        elif x == 1:
            path_obj['device'] = path_list[1]
        else:
            path_obj['tag_path'] = path_list[2].split('.')
    return path_obj

def _address_dedecimal(tag_address):
    '''Used to handle URL references where decimal notation isn't supported in object names, i.e. IoT Gateway items.
    
    Replaces `.` with `_` and removes leading `_` for system tag references'''
    if tag_address[0] == '_':
        tag_address = tag_address[1::]
    updated = tag_address.replace('.','_')
    return updated

def _url_parse_object(object):
    '''Common url parse to handle reserved characters. Used to convert object 
    names when building URLs.
    
    Reserved character list that Kepware allows in object names: :/?#[]@!$&'()*+,;='''
    return parse.quote(object, safe='')



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`iot_items` exposes an API to allow modifications (add, delete, modify) to 
iot_items objects within the Kepware Configuration API
"""

from typing import Union
from .. import utils
from ..connection import server
from .. import iot_gateway as IOT
from ..error import KepError, KepHTTPError
from ..utils import _url_parse_object

IOT_ITEMS_ROOT = '/iot_items'

def _create_url(tag = None):
    '''Creates url object for the "iot items" branch of Kepware's IoT Agents property model. Used 
    to build a part of Kepware Configuration API URL structure
    
    Returns the device specific url when a value is passed as the iot item name.
    '''
    if tag == None:
        return IOT_ITEMS_ROOT
    else: 
        normalized_tag = utils._address_dedecimal(tag)
        return '{}/{}'.format(IOT_ITEMS_ROOT, _url_parse_object(normalized_tag))


def add_iot_item(server: server, DATA: Union[dict, list], agent: str, agent_type: str) -> Union[bool, list]:
    '''Add a `"iot item"` or multiple `"iot item"` objects to Kepware's IoT Gateway agent. Additionally 
    it can be used to pass a list of iot items to be added to an agent all at once.

    :param server: instance of the `server` class
    :param DATA: Dict or List of Dicts of the iot item or list of items
    expected by Kepware Configuration API
    :param agent: name of IoT Agent
    :param agent_type: agent type. Valid values are `MQTT Client`, `REST Client` or `REST Server`

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    iot items added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_add(server.url + IOT.agent._create_url(agent_type, agent) + _create_url(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
            errors = [] 
            for item in r.payload:
                if item['code'] != 201:
                    errors.append(item)
            return errors 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_iot_item(server: server, iot_item: str, agent: str, agent_type: str) -> bool:
    '''Delete an `"iot item"` object in Kepware.

    :param server: instance of the `server` class
    :param iot_item: IoT item to delete
    :param agent: name of IoT Agent
    :param agent_type: agent type. Valid values are `MQTT Client`, `REST Client` or `REST Server`

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_del(server.url + IOT.agent._create_url(agent_type, agent) + _create_url(iot_item))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_iot_item(server: server, DATA: dict, agent: str, agent_type: str, *, iot_item: str = None, force: bool = False) -> bool:
    '''Modify an `"iot item"` object and it's properties in Kepware. If a `"iot item"` is not provided as an input,
    you need to identify the iot item in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the iot item that is to be modified.

    :param server: instance of the `server` class
    :param DATA: Dict of the iot item properties to be modified.
    :param agent: name of IoT Agent
    :param agent_type: agent type. Valid values are `MQTT Client`, `REST Client` or `REST Server`
    :param iot_item: *(optional)* name of IoT item to modify. Only needed if not existing in `"DATA"`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    agent_data = server._force_update_check(force, DATA)
    if iot_item == None:
        try:
            r = server._config_update(server.url + IOT.agent._create_url(agent_type, agent) + _create_url(agent_data['common.ALLTYPES_NAME']), agent_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No agent identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
    else:
        r = server._config_update(server.url + IOT.agent._create_url(agent_type, agent) + _create_url(iot_item), agent_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_iot_item(server: server, iot_item: str, agent: str, agent_type: str)-> dict:
    '''Returns the properties of the `"iot item"` object.

    :param server: instance of the `server` class
    :param iot_item: name of IoT item to retrieve properties
    :param agent: name of IoT Agent
    :param agent_type: agent type. Valid values are `MQTT Client`, `REST Client` or `REST Server`

    :return: Dict of properties for the iot item requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(server.url + IOT.agent._create_url(agent_type, agent) + _create_url(iot_item))
    return r.payload

def get_all_iot_items(server: server, agent: str, agent_type: str, *, options: dict = None) -> list:
    '''Returns the properties of all `"iot item"` objects for an agent.
    
    :param server: instance of the `server` class
    :param iot_item: name of IoT item to retrieve properties
    :param agent: name of IoT Agent
    :param agent_type: agent type. Valid values are `MQTT Client`, `REST Client` or `REST Server`
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of IoT items. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize'. Only used when exchange_name is not defined.

    :return: list of properties for all IoT items

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    r = server._config_get(f'{server.url}{IOT.agent._create_url(agent_type, agent)}{_create_url()}', params= options)
    return r.payload


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`admin` module provides functionality to manage Kepware Administration based 
properties available through the Kepware Configuration API
"""

from . import users, user_groups,ua_server, lls


# -------------------------------------------------------------------------
# Copyright (c), PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`user_groups` exposes an API to allow modifications (add, delete, modify) to 
user groups within the Kepware Administration User Manager through the Kepware Configuration API
"""
from typing import Union
from ..error import KepHTTPError, KepError
from ..connection import server
from ..utils import _url_parse_object


USERGROUPS_ROOT = '/admin/server_usergroups'
ENABLE_PROPERTY = 'libadminsettings.USERMANAGER_GROUP_ENABLED'

def _create_url(user_group = None):
    '''Creates url object for the "server_usergroups" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the user group specific url when a value is passed as the user_group name.
    '''
    
    if user_group == None:
        return USERGROUPS_ROOT
    else:
        return '{}/{}'.format(USERGROUPS_ROOT, _url_parse_object(user_group))

def add_user_group(server: server, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"user group"` or multiple `"user group"` objects to Kepware User Manager by passing a 
    list of user groups to be added all at once.

    :param server: instance of the `server` class
    :param DATA: Dict or List of Dicts of the user groups to add

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    endpoints added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(), DATA)
    if r.code == 201: return True 
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_user_group(server: server, user_group: str) -> bool:
    '''Delete a `"user group"` object in Kepware User Manager
    
    :param server: instance of the `server` class
    :param user_group: name of user group to delete
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(user_group))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_user_group(server: server, DATA: dict, *, user_group: str = None) -> bool:
    '''Modify a `"user group"` object and it's properties in Kepware User Manager. If a `"user group"` is not provided as an input,
    you need to identify the user group in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the user group that is to be modified.

    :param server: instance of the `server` class
    :param DATA: Dict of the user group properties to be modified.
    :param user_group: *(optional)* name of user group to modify. Only needed if not existing in `"DATA"`
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    if user_group == None:
        try:
            r = server._config_update(server.url + _create_url(DATA['common.ALLTYPES_NAME']), DATA)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No User Group identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
    else:
        r = server._config_update(server.url + _create_url(user_group), DATA)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_user_group(server: server, user_group: str) -> dict:
    '''Returns the properties of the `"user group"` object.
    
    :param server: instance of the `server` class
    :param user_group: name of user group to retrieve
    
    :return: Dict of properties for the user group requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_get(server.url + _create_url(user_group))
    return r.payload

def get_all_user_groups(server: server, *, options: dict = None) -> list:
    '''Returns list of all `"user group"` objects and their properties.
    
    :param server: instance of the `server` class
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of user groups. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize.

    :return: List of properties for all user groups

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_get(f'{server.url}{_create_url()}', params= options)
    return r.payload

def enable_user_group(server: server, user_group: str) -> bool:
    '''Enable the `"user group"`.
    
    :param server: instance of the `server` class
    :param user_group: name of user group

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: True}
    return modify_user_group(server, DATA, user_group= user_group)

def disable_user_group(server: server, user_group: str) -> bool:
    '''Disable the `"user group"`.
    
    :param server: instance of the `server` class
    :param user_group: name of user group

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: False}
    return modify_user_group(server, DATA, user_group= user_group)


# -------------------------------------------------------------------------
# Copyright (c), PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`users` exposes an API to allow modifications (add, delete, modify) to 
users within the Kepware Administration User Management through the Kepware Configuration API
"""
from typing import Union
from ..error import KepError, KepHTTPError
from ..connection import server
from ..utils import _url_parse_object


USERS_ROOT = '/admin/server_users'
ENABLE_PROPERTY = 'libadminsettings.USERMANAGER_USER_ENABLED'

def _create_url(user = None):
    '''Creates url object for the "server_users" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the user specific url when a value is passed as the user name.
    '''
    
    if user == None:
        return USERS_ROOT
    else:
        return '{}/{}'.format(USERS_ROOT, _url_parse_object(user))

def add_user(server: server, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"user"` or multiple `"user"` objects to Kepware User Manager by passing a 
    list of users to be added all at once.

    :param server: instance of the `server` class
    :param DATA: Dict or List of Dicts of the users to add

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    endpoints added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(), DATA)
    if r.code == 201: return True 
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_user(server: server, user: str) -> bool:
    '''Delete a `"user"` object in Kepware User Manager
    
    :param server: instance of the `server` class
    :param user: name of user to delete
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(user))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_user(server: server , DATA: dict, *, user: str = None) -> bool:
    '''Modify a `"user object"` and it's properties in Kepware User Manager. If a `"user"` is not provided as an input,
    you need to identify the user in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the user that is to be modified.

    :param server: instance of the `server` class
    :param DATA: Dict of the user properties to be modified.
    :param user: *(optional)* name of user to modify. Only needed if not existing in `"DATA"`
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    if user == None:
        try:
            r = server._config_update(server.url + _create_url(DATA['common.ALLTYPES_NAME']), DATA)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No User identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
    else:
        r = server._config_update(server.url + _create_url(user), DATA)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_user(server: server, user: str) -> dict:
    '''Returns the properties of the `"user"` object.
    
    :param server: instance of the `server` class
    :param user: name of user to retrieve
    
    :return: Dict of properties for the user requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_get(server.url + _create_url(user))
    return r.payload

def get_all_users(server: server, *, options: dict = None) -> list:
    '''Returns list of all `"user"` objects and their properties.
    
    :param server: instance of the `server` class
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of users. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize.

    :return: List of properties for all users

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_get(f'{server.url}{_create_url()}', params= options)
    return r.payload

def enable_user(server: server, user: str) -> bool:
    '''Enable the `"user"`.
    
    :param server: instance of the `server` class
    :param user: name of user

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: True}
    return modify_user(server, DATA, user= user)

def disable_user(server: server, user: str) -> bool:
    '''Disable the `"user"`.
    
    :param server: instance of the `server` class
    :param user: name of user

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: False}
    return modify_user(server, DATA, user= user)


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`exchange` exposes an API to allow modifications (add, delete, modify) to 
exchange objects for EGD devices within the Kepware Configuration API
"""

from ...connection import server
from ...error import KepHTTPError, KepError
from ...utils import _url_parse_object, path_split
from typing import Union
from .. import egd as EGD, channel, device

CONSUMER_ROOT = '/consumer_exchange_groups/consumer exchanges/consumer_exchanges'
PRODUCER_ROOT = '/producer_exchange_groups/producer exchanges/producer_exchanges'

def _create_url(device_path, ex_type, exchange_name = None):
    '''Creates url object for the "exchange" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the exchange specific url when a value is passed as the exchange name.
    '''
    path_obj = path_split(device_path)
    device_root = channel._create_url(path_obj['channel']) + device._create_url(path_obj['device'])

    if exchange_name == None:
        if ex_type.upper() == EGD.CONSUMER_EXCHANGE:
            return device_root + CONSUMER_ROOT
        else:
            return device_root + PRODUCER_ROOT
    else:
        if ex_type.upper() == EGD.CONSUMER_EXCHANGE:
            return '{}{}/{}'.format(device_root,CONSUMER_ROOT,_url_parse_object(exchange_name))
        else:
            return '{}{}/{}'.format(device_root,PRODUCER_ROOT,_url_parse_object(exchange_name))

def add_exchange(server: server, device_path: str, ex_type: str, DATA: Union[dict, list]) -> Union[bool, list]:
    '''Add a `"exchange"` or multiple `"exchange"` objects to Kepware. Can be used to pass children of a exchange object 
    such as ranges. This allows you to create a exchange and ranges for the exchange all in one function, if desired.

    Additionally it can be used to pass a list of exchanges and it's children to be added all at once.

    :param server: instance of the `server` class
    :param device_path: path to EGD device with exchanges. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param ex_type: type of exchange, either `CONSUMER` or `PRODUCER`
    :param DATA: Dict or List of Dicts of the exchange(s) and it's children
    expected by Kepware Configuration API

    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    exchanges added that failed.
        
    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(device_path, ex_type), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_exchange(server: server, device_path: str, ex_type: str, exchange_name: str) -> bool:
    '''Delete an `"exchange"` object in Kepware. This will delete all children as well
    
    :param server: instance of the `server` class
    :param device_path: path to EGD device with exchanges. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param ex_type: type of exchange, either `CONSUMER` or `PRODUCER`
    :param exchange_name: name of exchange to delete
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(device_path, ex_type, exchange_name))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_exchange(server: server, device_path: str, ex_type: str, DATA: dict, *, exchange_name: str = None, force: bool = False) -> bool:
    '''Modify a `"exchange"` object and it's properties in Kepware. If a `"exchange_name"` is not provided as an input,
    you need to identify the exchange in the *'common.ALLTYPES_NAME'* property field in the `"DATA"`. It will 
    assume that is the exchange that is to be modified.

    :param server: instance of the `server` class
    :param device_path: path to EGD device with exchanges. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param DATA: Dict of the exchange properties to be modified.
    :param ex_type: type of exchange, either `CONSUMER` or `PRODUCER`
    :param exchange_name: *(optional)* name of exchange to modify. Only needed if not existing in `"DATA"`
    :param force: *(optional)* if True, will force the configuration update to the Kepware server
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    
    exchange_data = server._force_update_check(force, DATA)
    if exchange_name == None:
        try:
            r = server._config_update(server.url + _create_url(device_path, ex_type, exchange_data['common.ALLTYPES_NAME']), exchange_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = f'Error: No exchange identified in DATA | Key Error: {type(DATA)}'
            raise KepError(err_msg) 
    else:
        r = server._config_update(server.url + _create_url(device_path, ex_type, exchange_name), exchange_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_exchange(server: server, device_path: str, ex_type: str, exchange_name: str = None, *, options: dict = None) -> Union[dict, list]:
    '''Returns the properties of the exchange object or a list of all exchanges and their 
    properties for the type input. Returned object is JSON.
    
    :param server: instance of the `server` class
    :param device_path: path to EGD device with exchanges. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param ex_type: type of exchange, either `CONSUMER` or `PRODUCER`
    :param exchange_name: *(optional)* name of exchange. If not defined, get all exchanges
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of exchanges. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize'. Only used when exchange_name is not defined.
    
    :return: Dict of properties for the exchange requested or a List of exchanges and their properties

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    if exchange_name == None:
        r = server._config_get(f'{server.url}{_create_url(device_path, ex_type)}', params= options)
    else:
        r = server._config_get(f'{server.url}{_create_url(device_path, ex_type, exchange_name)}')
    return r.payload

def get_all_exchanges(server: server, device_path: str, *, options: dict = None) -> list[list, list]:
    '''Returns list of all `"exchange"` objects (both CONSUMER and PRODUCER) and their properties. Returned object is JSON list.
    
    INPUTS:
    :param server: instance of the `server` class
    :param device_path: path to EGD device with exchanges. Standard Kepware address decimal 
    notation string such as `"channel1.device1"`
    :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of exchanges. Options are 'filter', 
    'sortOrder', 'sortProperty', 'pageNumber', and 'pageSize'. Only used when exchange_name is not defined.
    
    :return: List - [list of consumer exchanges, list of producer exchanges] - list of lists for all 
    exchanges for the device

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    exchange_list = []
    exchange_list.append(get_exchange(server, device_path, EGD.CONSUMER_EXCHANGE, options= options))
    exchange_list.append(get_exchange(server, device_path, EGD.PRODUCER_EXCHANGE, options= options))
    return exchange_list


# filepath: c:\Users\rlabbe\OneDrive - PTC\Documents\GitHub\Kepware-ConfigAPI-SDK-Python\kepconfig\connectivity\adv_tag_group.py
# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# Note: The code within this file was created in total or in part
#  with the use of AI tools.


r"""`adv_tag_group` exposes an API to allow modifications (add, delete, modify) to 
tag group objects within the Kepware Configuration API
"""

from ..connection import KepServiceResponse, server
from ..error import KepHTTPError, KepError
from ..utils import _url_parse_object, path_split
from typing import Union
from .. import adv_tags
import inspect

TAG_GROUP_ROOT = '/advanced_tag_groups'

def _create_url(tag_group=None):
    '''Creates url object for the "tag group" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure
    
    Returns the tag group specific url when a value is passed as the tag group name.
    '''
    if tag_group is None:
        return TAG_GROUP_ROOT
    else:
        return '{}/{}'.format(TAG_GROUP_ROOT, _url_parse_object(tag_group))

def _create_adv_tags_group_url(path_obj):
    '''Creates url object for the "path_obj" which provides the adv tags tag group structure of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure
    
    Returns the advanced tag group specific url when a value is passed as the tag group name.
    '''

    try:
        url = ''
        if 'tag_path' in path_obj:
            for obj in path_obj['tag_path']:
                url += _create_url(obj)
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = f'Error: Error with {inspect.currentframe().f_code.co_name}: {str(e)}'
        raise KepError(err_msg)
    return url

def add_tag_group(server: server, adv_tag_group_path: str, DATA: Union[dict, list]) -> Union[bool, list]:
    # TODO: confirm adding tag types in this folder
    # TODO: Do we need to require the tag group path if we are adding a tag group at the root? (i.e. _advancedtags)
    '''Add a `"tag group"` or multiple `"tag group"` objects to a device in Kepware. Can be used to pass children of a tag group object 
    such as tags. This allows you to create a tag group and tags 
    all in one function, if desired.

    Additionally it can be used to pass a list of tag groups and its children to be added all at once.

    :param server: instance of the `server` class
    :param adv_tag_group_path: path identifying tag group to add the tag group object(s). Standard Kepware address decimal 
    notation string such as "_advancedtags.AdvTagGroup1" or "_advancedtags.AdvTagGroup1.AdvTagGroupChild"
    :param DATA: Dict or List of Dicts of the tag group(s) and its children
    expected by Kepware Configuration API
    
    :return: True - If a "HTTP 201 - Created" is received from Kepware server
    :return: If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    tag groups added that failed.

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    path_obj = adv_tags._adv_tag_path_split(adv_tag_group_path, isItem=False)
    url = adv_tags._create_adv_tags_base_url(server.url, path_obj) + _create_url()
    
    r = server._config_add(url, DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = []
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else:
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_tag_group(server: server, adv_tag_group_path: str) -> bool:
    '''Delete a `"tag group"` object in Kepware. This will delete all children as well.

    :param server: instance of the `server` class
    :param adv_tag_group_path: path identifying tag group to delete. Standard Kepware address decimal 
    notation string such as "_advancedtags.AdvTagGroup1" or "_advancedtags.AdvTagGroup1.AdvTagGroupChild"

    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    # url = _create_url(server.url, adv_tag_group_path, isItem=True)
    path_obj = adv_tags._adv_tag_path_split(adv_tag_group_path, isItem=True)
    url = adv_tags._create_adv_tags_base_url(server.url, path_obj) + _create_url(path_obj['item'])
    r = server._config_del(url)
    if r.code == 200: return True
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_tag_group(server: server, adv_tag_group_path: str, DATA: dict, force: bool = False) -> bool:
    '''Modify a tag group object and its properties in Kepware.

    :param server: instance of the `server` class
    :param adv_tag_group_path: path identifying the advanced tag group to modify. Standard Kepware address decimal 
    notation string such as "_advancedtags.AdvTagGroup1" or "_advancedtags.AdvTagGroup1.AdvTagGroupChild"
    :param DATA: Dict of the advanced tag group properties to be modified
    :param force: *(optional)* if True, will force the configuration update to the Kepware server
    
    :return: True - If a "HTTP 200 - OK" is received from Kepware server

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''

    tag_group_data = server._force_update_check(force, DATA)
    path_obj = adv_tags._adv_tag_path_split(adv_tag_group_path, isItem=True)
    url = adv_tags._create_adv_tags_base_url(server.url, path_obj) + _create_url(path_obj['item'])

    r = server._config_update(url, tag_group_data)
    if r.code == 200: return True
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_tag_group(server: server, adv_tag_group_path: str) -> dict:
    '''Returns the properties of the tag group object.

    :param server: instance of the `server` class
    :param adv_tag_group_path: path identifying the advanced tag group to retrieve. Standard Kepware address decimal
    notation string such as "_advancedtags.AdvTagGroup1" or "_advancedtags.AdvTagGroup1.AdvTagGroupChild"

    :return: Dict of data for the tag group requested

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    path_obj = adv_tags._adv_tag_path_split(adv_tag_group_path, isItem=True)
    url = adv_tags._create_adv_tags_base_url(server.url, path_obj) + _create_url(path_obj['item'])
    r = server._config_get(url)
    return r.payload

def get_all_tag_groups(server: server, adv_tag_group_path: str, *, options: dict = None) -> list:
    '''Returns list of all tag group objects and their properties within a tag group. Returned object is JSON list.
    
    :param server: instance of the `server` class
    :param adv_tag_group_path: path identifying the advanced tag group collection to retrieve. Standard Kepware address decimal
    notation string such as "_advancedtags.AdvTagGroup1" or "_advancedtags.AdvTagGroup1.AdvTagGroupChild"

    :return: List of data for all tag groups within the tag group

    :raises KepHTTPError: If urllib provides an HTTPError
    :raises KepURLError: If urllib provides an URLError
    '''
    path_obj = adv_tags._adv_tag_path_split(adv_tag_group_path, isItem=True)
    url = adv_tags._create_adv_tags_base_url(server.url, path_obj) + _create_url()
    r = server._config_get(url, params=options)
    return r.payload



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`connection` exposes an `server` class that manages the connection
information and RESTful requests for the Kepware Configuration API Library.
"""

import json
import codecs
import datetime
from urllib import request, parse, error
from base64 import b64encode
from .error import KepError, KepHTTPError, KepURLError
import socket
import ssl
from .structures import KepServiceResponse, KepServiceStatus, _HttpDataAbstract, Filter


class server:
    '''A class to represent a connection to an instance of Kepware. This object is used to 
    create the Authentication and server parameters to taget a Kepware instance. An instance of this is 
    used in all configuration calls done.

    :param host: host name or IP address
    :param port: port of Configuration API
    :param username: username to conduct "Basic Authentication"
    :param password: password to conduct "Basic Authentication"
    :param https: Sets `SSL_on` to use HTTPS connection (Default: False)
    :param SSL_on: Identify to use HTTPS connection (Default: False)
    :param SSL_ignore_hostname: During certificate validation ignore the hostname check
    :param SSL_trust_all_certs: (insecure) - During certificate validation trust any certificate - if True, 
        will "set SSL_ignore_hostname" to true
    :param url: base URL for the server connection

    **Methods**

    :meth:`reinitialize`: reinitialize the Kepware server

    :meth:`get_transaction_log`: retrieve the Configuration API transaction logs

    :meth:`get_event_log`: retrieve the Kepware Event Log

    :meth:`get_audit_log`: retrieve the Kepware Audit Log

    :meth:`get_info`: retrieve the Kepware product information
    
    :meth:`import_empty_project`: import an empty project to the Kepware server

    :meth:`get_project_properties`: retrieve the Kepware Project Properties

    :meth:`modify_project_properties` - modify the Kepware Project Properties

    :meth:`service_status` - retrive service job status

    :meth:`export_project_configuration` - export the current project configuration in JSON format

    :meth:`save_project` - save the current project to a file

    :meth:`load_project` - load a project from a file
    '''
    __root_url = '/config'
    __version_url = '/v1'
    __project_services_url = '/project/services'
    __event_log_url = '/event_log'
    __trans_log_url = '/log'
    __audit_log_url = '/audit_log'



    def __init__(self,  host: str, port: int, user: str, pw: str, https: bool = False):
        self.host = host
        self.port = port
        self.username = user
        self.password = pw
        self.__ssl_context = ssl.create_default_context()
        self.__SSL_on = https
    
    @property
    def url(self):
        if self.SSL_on:
            proto = 'https'
        else:
            proto = 'http'
        return  f'{proto}://{self.host}:{self.port}{self.__root_url}{self.__version_url}'
    
    @property
    def SSL_on(self):
        return self.__SSL_on
    
    @SSL_on.setter
    def SSL_on(self, val):
        
        if isinstance(val, bool):
            self.__SSL_on = val

    @property
    def SSL_ignore_hostname(self):
        return not self.__ssl_context.check_hostname

    @SSL_ignore_hostname.setter
    def SSL_ignore_hostname(self, val):
        if isinstance(val, bool):
            if val == True:
                self.__ssl_context.check_hostname = False
            else:
                self.__ssl_context.check_hostname = True

    
    @property
    def SSL_trust_all_certs(self):
        if self.__ssl_context.verify_mode == ssl.CERT_NONE:
            return True
        else:
            return False

    @SSL_trust_all_certs.setter
    def SSL_trust_all_certs(self, val):
        if isinstance(val, bool):
            if val == True:
                if self.__ssl_context.check_hostname == True:
                    self.__ssl_context.check_hostname = False
                self.__ssl_context.verify_mode = ssl.CERT_NONE
            else:
                self.__ssl_context.verify_mode = ssl.CERT_REQUIRED


    def get_status(self) -> dict:
        '''Executes a health status request to the Kepware instance to report service statuses.

        :return: List of data for the health status request

        :raises KepHTTPError: If urllib provides an HTTPError
        :raises KepURLError: If urllib provides an URLError
        '''

        r = self._config_get(f'{self.url}/status')
        return r.payload
    def get_info(self) -> dict:
        '''Requests product information from the Kepware instance. Provides various information including
        product name and version information.

        :return: dict of data for the product information request
        
        :raises KepHTTPError: If urllib provides an HTTPError
        :raises KepURLError: If urllib provides an URLError
        '''
        
        r = self._config_get(f'{self.url}/about')
        return r.payload

    def reinitialize(self, job_ttl: int = None) -> KepServiceResponse:
        '''Executes a Reinitialize service call to the Kepware instance.

        :param job_ttl: *(optional)* Determines the number of seconds a job instance will exist following completion.

        :return: `KepServiceResponse` instance with job information

        :raises KepHTTPError: If urllib provides an HTTPError (If not HTTP code 202 [Accepted] or 429 [Too Busy] returned)
        :raises KepURLError: If urllib provides an URLError
        '''
        url = self.url + self.__project_services_url + '/ReinitializeRuntime' 
        try:
            job = self._kep_service_execute(url, None, job_ttl)
            return job
        except Exception as err:
            raise err
        
    def get_transaction_log(self, limit: int = None, start: datetime.datetime = None, end: datetime.datetime = None) -> list:
        ''' Get the Transaction Log from the Kepware instance.

        :param limit: *(optional)* number of transaction log entries to request
        :param start: *(optional)* start time of query as `datetime.datetime` type and should be UTC
        :param end: *(optional)* end time of query as `datetime.datetime` type and should be UTC

        :raises KepHTTPError: If urllib provides an HTTPError
        :raises KepURLError: If urllib provides an URLError
        '''
        query = self.__create_query(start, end, limit)
        url = f'{self.url}{self.__trans_log_url}'
        r = self._config_get(url, params= query)
        return r.payload

    def get_event_log(self, limit: int = None, start: datetime.datetime = None, end: datetime.datetime = None, *, options: dict = None) -> list:
        ''' Get the Event Log from the Kepware instance.

        :param limit: *(optional)* number of event log entries to request
        :param start: *(optional)* start time of query as `datetime.datetime` type and should be UTC
        :param end: *(optional)* end time of query as `datetime.datetime` type and should be UTC
        :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of transactions. Options are `event`, 
        `sortOrder`, `sortProperty`, `pageNumber`, and `pageSize`

        :raises KepHTTPError: If urllib provides an HTTPError
        :raises KepURLError: If urllib provides an URLError
        '''
        query = self.__create_query(start, end, limit)
        if options is not None:
            query = {**query, **options}
        url = f'{self.url}{self.__event_log_url}'
        r = self._config_get(url, params= query)
        return r.payload

    def get_audit_log(self, limit: int = None, *, filters: list[Filter] = None, options: dict = None) -> list:
        ''' Get the Audit Log from the Kepware instance.

        :param limit: *(optional)* number of event log entries to request
        :param filters: *(optional)* list of filters that are used to control results returned from the log
        :param options: *(optional)* Dict of parameters to filter, sort or pagenate the list of transactions. Options are `sortOrder`, 
        `sortProperty`, `pageNumber`, and `pageSize`

        :raises KepHTTPError: If urllib provides an HTTPError
        :raises KepURLError: If urllib provides an URLError
        '''
        query = self.__create_filter_query(filters)
        if limit is not None:
            query['limit'] = limit
        if options is not None:
            query = {**query, **options}
        url = f'{self.url}{self.__audit_log_url}'
        r = self._config_get(url, params= query)
        return r.payload
    
    def get_project_properties(self) -> dict:
        ''' Get the Project Properties of the Kepware instance.
        
        :return: Dict of all the project properties

        :raises KepHTTPError: If urllib provides an HTTPError
        :raises KepURLError: If urllib provides an URLError
        '''

        r = self._config_get(self.url + '/project')
        return r.payload
    
    def modify_project_properties(self, DATA: dict, force: bool = False) -> bool:
        ''' Modify the Project Properties of the Kepware instance.

        :param DATA: Dict of the project properties to be modified
        :param force: *(optional)* if True, will force the configuration update to the Kepware server
        
        :return: True - If a "HTTP 200 - OK" is received from Kepware server

        :raises KepHTTPError: If urllib provides an HTTPError
        :raises KepURLError: If urllib provides an URLError
        '''

        prop_data = self._force_update_check(force, DATA)
        r = self._config_update(self.url + '/project', prop_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
    
    def import_empty_project(self) -> KepServiceResponse:
        '''Executes JsonProjectLoad Service call to the Kepware instance with an empty project. This service 
        imports an empty project configuration, acts like a FILE->NEW action and 
        stop communications while the new project replaces the current project in the Kepware runtime. 

        :return: `KepServiceResponse` instance with job information
        
        :raises KepHTTPError: If urllib provides an HTTPError (If not HTTP code 202 [Accepted] or 429 [Too Busy] returned)
        :raises KepURLError: If urllib provides an URLError
        '''
        return self.import_project_configuration({"project":{}})


    def import_project_configuration(self, DATA: dict) -> KepServiceResponse:
        '''Executes JsonProjectLoad Service call to the Kepware instance. This service imports project configuration 
        data, expecting a complete project file in JSON/dict format. This service acts like a FILE->OPEN action and 
        stop communications while the new project replaces the current project in the Kepware runtime. 
    
        :param DATA: Complete project configuration data in JSON/dict format. 

        :return: `KepServiceResponse` instance with job information
        
        :raises KepHTTPError: If urllib provides an HTTPError (If not HTTP code 202 [Accepted] or 429 [Too Busy] returned)
        :raises KepURLError: If urllib provides an URLError
        '''
        url = self.url + self.__project_services_url + '/JsonProjectLoad'
        try:
            job = self._kep_service_execute(url, DATA)
            return job
        except Exception as err:
            raise err
        
    def export_project_configuration(self) -> dict:
        '''Get a complete copy of the project configuration in JSON format. This will include the same 
        configuration that is stored when you save the project file manually.

        :return: Dict of the complete project configuration

        :raises KepHTTPError: If urllib provides an HTTPError
        :raises KepURLError: If urllib provides an URLError
        '''
        r = self._config_get(self.url + '/project', params= {"content": "serialize"})
        return r.payload
    
    def save_project(self, filename: str, password: str = None, job_ttl: int = None) -> KepServiceResponse:
        '''Executes a ProjectSave Service call to the Kepware instance. This saves 
        a copy of the current project file to disk. The filename

        :param filename: Relative file path and name of project file including the file extension to save.
        Location of relative project file paths:
         
                TKS or KEP (Windows): C:\\PROGRAMDATA\\PTC\\Thingworx Kepware Server\\V6 or 
                                C:\\PROGRAMDATA\\Kepware\\KEPServerEX\\V6
                TKE (Linux):    /opt/tkedge/v1/user_data


        :param password: *(optional)* Specify a password with which to  save an encrypted project file with.  
            This password will be required to load this project file.
        :param job_ttl: *(optional)* Determines the number of seconds a job instance will exist following completion.

        :return: `KepServiceResponse` instance with job information

        :raises KepHTTPError: If urllib provides an HTTPError (If not HTTP code 202 [Accepted] or 429 [Too Busy] returned)
        :raises KepURLError: If urllib provides an URLError
        '''
        url = self.url + self.__project_services_url + '/ProjectSave'
        prop_data = {'servermain.PROJECT_FILENAME': filename}
        if password != None: prop_data['servermain.PROJECT_PASSWORD'] = password
        try:
            job = self._kep_service_execute(url, prop_data, job_ttl)
            return job
        except Exception as err:
            raise err

    def load_project(self, filename: str, password: str = None, job_ttl: int = None) -> KepServiceResponse:
        '''Executes a ProjectLoad Service call to the Kepware instance. This loads 
        a project file to disk.

        INPUTS:
        
        :param filename: Fully qualified or relative path and name of project file including the file extension. Absolute
        paths required for TKS and KEP while file path is relative for TKE:

                Windows - filename = C:\\filepath\\test.opf
                Linux - filename = filepath/test.lpf - Location is /opt/tkedge/v1/user_data/filepath/test.lpf

        :param password: *(optional)* Specify a password with which to load an encrypted project file.          
        :param job_ttl: *(optional)* Determines the number of seconds a job instance will exist following completion.

        :return: `KepServiceResponse` instance with job information

        :raises KepHTTPError: If urllib provides an HTTPError (If not HTTP code 202 [Accepted] or 429 [Too Busy] returned)
        :raises KepURLError: If urllib provides an URLError
        '''
        url = self.url + self.__project_services_url + '/ProjectLoad'
        prop_data = {'servermain.PROJECT_FILENAME': filename}
        if password != None: prop_data['servermain.PROJECT_PASSWORD'] = password
        try:
            job = self._kep_service_execute(url, prop_data, job_ttl)
            return job
        except Exception as err:
            raise err

    def get_project_backup_info(self) -> dict:
        ''' Get the Project Backup Information of the Kepware instance.
        
        :return: List of all the backup projects and their properties.

        :raises KepHTTPError: If urllib provides an HTTPError
        :raises KepURLError: If urllib provides an URLError
        '''

        r = self._config_get(self.url + '/project/backups')
        return r.payload

    def backup_project(self, job_ttl: int = None) -> KepServiceResponse:
        '''Executes a CreateBackup Service call to the Kepware instance. This saves 
        a copy of the current project file to disk as a backup that can be retrieved.

        :param job_ttl: *(optional)* Determines the number of seconds a job instance will exist following completion.

        :return: `KepServiceResponse` instance with job information

        :raises KepHTTPError: If urllib provides an HTTPError (If not HTTP code 202 [Accepted] or 429 [Too Busy] returned)
        :raises KepURLError: If urllib provides an URLError
        '''
        url = self.url + self.__project_services_url + '/CreateBackup'
        try:
            job = self._kep_service_execute(url, TTL= job_ttl)
            return job
        except Exception as err:
            raise err
        
    def service_status(self, resp: KepServiceResponse):
        '''Returns the status of a service job. Used to verify if a service call
        has completed or not.

        :param resp: `KepServiceResponse` instance with job information

        :return: `KepServiceStatus` instance with job status

        :raises KepHTTPError: If urllib provides an HTTPError
        :raises KepURLError: If urllib provides an URLError
        '''
        # need to remove part of job href
        loc = resp.href.find(self.__root_url + self.__version_url)
        job_url = resp.href[loc + len(self.__root_url + self.__version_url):]

        r = self._config_get(self.url + job_url)
        job = KepServiceStatus(r.payload['servermain.JOB_COMPLETE'],r.payload['servermain.JOB_STATUS'], r.payload['servermain.JOB_STATUS_MSG'])
        return job


    #Function used to Add an object to Kepware (HTTP POST)
    def _config_add(self, url, DATA):
        '''Conducts an POST method at *url* to add an object in the Kepware Configuration
        *DATA* is required to be a properly JSON object (dict) of the item to be posted to *url* 
        '''
        if len(DATA) == 0:
            err_msg = f'Error: Empty List or Dict in DATA | DATA type: {type(DATA)}'
            raise KepError(err_msg) 
        data = json.dumps(DATA).encode('utf-8')
        url_obj = self.__url_validate(url)
        q = request.Request(url_obj, data, method='POST')
        r = self.__connect(q)
        return r

    #Function used to del an object to Kepware (HTTP DELETE)
    def _config_del(self, url):
        '''Conducts an DELETE method at *url* to delete an object in the Kepware Configuration'''
        url_obj = self.__url_validate(url)
        q = request.Request(url_obj, method='DELETE')
        r = self.__connect(q)
        return r

    #Function used to Update an object to Kepware (HTTP PUT)
    def _config_update(self, url, DATA = None):
        '''Conducts an PUT method at *url* to modify an object in the Kepware Configuration.
        *DATA* is required to be a properly JSON object (dict) of the item to be put to *url*
        '''
        url_obj = self.__url_validate(url)
        if DATA == None:            
            q = request.Request(url_obj, method='PUT')
        else:
            data = json.dumps(DATA).encode('utf-8')
            q = request.Request(url_obj, data, method='PUT')
        r = self.__connect(q)
        return r

    #Function used to Read an object from Kepware (HTTP GET) and return the JSON response
    def _config_get(self, url, *, params = None):
        '''
        Conducts an GET method at *url* to retrieve an objects properties with query parameters in 
        the Kepware Configuration.
        '''
        # Add parameters when necessary
        if params is not None and params != {}:
            qparams = parse.urlencode(params)
            url = f'{url}?{qparams}'
        url_obj = self.__url_validate(url)
        q = request.Request(url_obj, method='GET')
        r = self.__connect(q)
        return r

    
    def _force_update_check(self, force, DATA):
        '''
        This will validate if the modify call needs to be forced or not. If forced, the dict DATA needs
        to have the 'FORCE_UPDATE' property with a value of True. If forced is not requested, it is necessary
        to provide the current 'PROJECT_ID'. If 'PROJECT_ID' is not present in DATA, this will automatically 
        retreive it from the active server.
        '''
        if force == True:
            DATA['FORCE_UPDATE'] = True
        else:
            # Get Project ID if it doesn't exist and if FORCE_UPDATE is existing and FALSE
            if 'PROJECT_ID' not in DATA:
                if 'FORCE_UPDATE' in DATA:
                    if 'FORCE_UPDATE' == False:
                        try:
                            project_data = self._config_get(self.url + '/project')
                            DATA['PROJECT_ID'] = project_data.payload['PROJECT_ID']
                        except:
                            #NEED TO COVER ERROR CONDITION
                            pass
                else:
                    try:
                        project_data = self._config_get(self.url + '/project')
                        DATA['PROJECT_ID'] = project_data.payload['PROJECT_ID']
                    except:
                        #NEED TO COVER ERROR CONDITION
                        pass
        return DATA
    # General service call handler
    def _kep_service_execute(self, url, DATA = None, TTL = None):
        try:
            if TTL != None:
                if DATA == None: DATA = {}
                DATA["servermain.JOB_TIME_TO_LIVE_SECONDS"]= TTL
            r = self._config_update(url, DATA)
            job = KepServiceResponse(r.payload['code'],r.payload['message'], r.payload['href'])
            return job
        except KepHTTPError as err:
            if err.code == 429:
                job = KepServiceResponse()
                job.code = err.code
                job.message = err.payload
                return job
            else:
                raise err

# 
# Supporting Functions
#

    # General connect call to manage HTTP responses for all methods
    # Returns the response object for the method to handle as appropriate
    # Raises Errors as found
    def __connect(self,request_obj):
        # Fill appropriate header information
        data = _HttpDataAbstract()
        request_obj.add_header("Authorization", "Basic %s" % self.__build_auth_str(self.username, self.password))
        request_obj.add_header("Content-Type", "application/json")
        request_obj.add_header("Accept", "application/json")
        try:
            # context is sent regardless of HTTP or HTTPS - seems to be ignored if HTTP URL
            with request.urlopen(request_obj, context=self.__ssl_context) as server:
                try:
                    payload = server.read()
                    data.payload = json.loads(codecs.decode(payload,'utf-8-sig'))
                except:
                    pass
                data.code = server.code
                data.reason = server.reason
                return data
        except error.HTTPError as err:
            payload = json.loads(codecs.decode(err.read(),'utf-8-sig'))
            # print('HTTP Code: {}\n{}'.format(err.code,payload), file=sys.stderr)
            raise KepHTTPError(url=err.url, code=err.code, msg=err.msg, hdrs=err.hdrs, payload=payload)
        except error.URLError as err:
            # print('URLError: {} URL: {}'.format(err.reason, request_obj.get_full_url()), file=sys.stderr)
            raise KepURLError(msg=err.reason, url=request_obj.get_full_url())

    # Fucntion used to ensure special characters are handled in the URL
    # Ex: Space will be turned to %20
    def __url_validate(self, url):
        # Configuration API does not use fragments in URL so ignore to allow # as a character
        # Objects in Kepware can include # as part of the object names
        parsed_url = parse.urlparse(url, allow_fragments= False)
        # Added % for scenarios where special characters have already been escaped with %
        updated_path = parse.quote(parsed_url.path, safe = '/%')

        # If host is "localhost", force using the IPv4 loopback adapter IP address in all calls
        # This is done to remove retries that will happen when the host resolution uses IPv6 intially
        # Kepware currently doesn't support IPv6 and is not listening on this interface
        if parsed_url.hostname.lower() == 'localhost':
            ip = socket.gethostbyname(parsed_url.hostname)
            parsed_url = parsed_url._replace(netloc='{}:{}'.format(ip, parsed_url.port))
        
        return parsed_url._replace(path=updated_path).geturl()

    # Function used to build the basic authentication string
    def __build_auth_str(self, username, password):
        if isinstance(username, str):
            username = username.encode('latin1')
        if isinstance(password, str):
            password = password.encode('latin1')
        authstr = b64encode(b':'.join((username, password))).strip().decode('ascii')
        return authstr
    
    # Create parameters for log queries
    def __create_query(self, start = None, end = None, limit = None):
        query = {}
        if start != None and isinstance(start, datetime.datetime):
            query['start'] = start.isoformat()
        if end != None and isinstance(end, datetime.datetime):
            query['end'] = end.isoformat()
        if limit != None:
            query['limit'] = limit
        return query

    # Create filter query for log queries
    def __create_filter_query(self, filters: list[Filter] = None):
        query = {}
        if filters is None:
            return query
        for f in filters:
            # Ensure we use the value of the Enum, not the Enum object itself
            key = f"filter[{f.field.value}][{f.modifier.value}]"
            query[key] = f.value
        return query